#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import abc
import importlib
import json
import logging
import pkgutil
from typing import Tuple, Union, List, Optional

from urllib.parse import urlparse
from ..params import KeeperParams
from .. import config_storage, utils

CONFIG_STORAGE_URL = 'config_storage'
PROTECTED_PROPERTIES = ['server', 'device_token', ('private_key', 'device_private_key')]   # type: List[Union[str, Tuple[str, str]]]
PROTECTED_CONNECTED_PROPERTIES = ['user', 'clone_code']   # type: List[Union[str, Tuple[str, str]]]
PROTECTED_READONLY_PROPERTIES = ['password']
EDITABLE_PROPERTIES = ['proxy']
BOOL_PROPERTIES = ['debug', 'batch_mode', 'unmask_all']
INT_PROPERTIES = ['timedelay', 'logout_timer']
ENCRYPTED_DATA = 'encrypted_data'


class SecureStorageException(Exception):
    pass


class SecureStorageBase(abc.ABC):
    @abc.abstractmethod
    def load_configuration(self, url, encrypted_data=None):   # type: (str, Optional[bytes]) -> dict
        pass

    @abc.abstractmethod
    def store_configuration(self, url, configuration):   # type: (str, dict) -> Optional[bytes]
        pass


def _get_plugin(url):  # type: (str) -> SecureStorageBase
    if not url:
        raise SecureStorageException(f'Configuration file error: "{CONFIG_STORAGE_URL}" cannot be empty')
    par = urlparse(url)
    if not par.scheme:
        raise SecureStorageException(f'Configuration file error: "{CONFIG_STORAGE_URL}" is not URL')

    plugin_name = par.scheme
    if not any(x for x in pkgutil.iter_modules(config_storage.__path__) if x.name == plugin_name):
        raise SecureStorageException(f'Protected storage "{plugin_name}" is not supported')

    module_name = f'keepercommander.config_storage.{plugin_name}'
    try:
        logging.debug('Importing config storage module "%s"', module_name)
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as me:
        dependency_not_installed = f'{plugin_name} requires {me.name} package to be installed\n' \
                                   f'pip install {me.name}'
        raise SecureStorageException(dependency_not_installed)
    if not hasattr(module, 'SecureStorage'):
        raise SecureStorageException(f'Protected storage "{plugin_name}" is invalid')
    storage_class = getattr(module, 'SecureStorage')
    storage = storage_class()     # type: SecureStorageBase
    if not isinstance(storage, SecureStorageBase):
        raise SecureStorageException(f'Protected storage "{plugin_name}" is invalid')
    return storage


def split_name(name):   # type: (Union[str, Tuple[str, str]]) -> Tuple[str, str]
    if isinstance(name, tuple):
        return name
    return name, name


def store_config_properties(params):
    if not isinstance(params, KeeperParams):
        return
    if not isinstance(params.config, dict):
        params.config = {}

    protected_properties = PROTECTED_PROPERTIES
    if params.session_token:
        protected_properties += PROTECTED_CONNECTED_PROPERTIES
    # commit changes from params to config
    for name in protected_properties + EDITABLE_PROPERTIES:
        config_name, params_name = split_name(name)
        if hasattr(params, params_name):
            value = getattr(params, params_name)
            if value:
                params.config[config_name] = value
            elif config_name in params.config:
                del params.config[config_name]

    config_json = params.config.copy()

    if CONFIG_STORAGE_URL in config_json:
        url = config_json[CONFIG_STORAGE_URL]
        storage = _get_plugin(url)
        conf_protected = {}
        for name in protected_properties + PROTECTED_READONLY_PROPERTIES:
            config_name, _ = split_name(name)
            if config_name in config_json:
                value = config_json[config_name] or ''
                if value:
                    conf_protected[config_name] = value
                del config_json[config_name]

        encrypted_data = storage.store_configuration(url, conf_protected)
        if isinstance(encrypted_data, bytes):
            config_json[ENCRYPTED_DATA] = utils.base64_url_encode(encrypted_data)

    if params.config_filename:
        try:
            with open(params.config_filename, 'w') as fd:
                json.dump(config_json, fd, ensure_ascii=False, indent=2)
        except Exception as error:
            logging.debug(error, exc_info=True)
            logging.error(f'Failed to write configuration to {params.config_filename}. '
                          'Type "debug" to toggle verbose logging.', error)


def load_config_properties(params):
    if not isinstance(params, KeeperParams):
        return

    if not isinstance(params.config, dict):
        return

    if CONFIG_STORAGE_URL in params.config:
        url = params.config[CONFIG_STORAGE_URL]
        storage = _get_plugin(url)
        encrypted_data = None
        if ENCRYPTED_DATA in params.config:
            ed = params.config[ENCRYPTED_DATA]
            if isinstance(ed, str):
                encrypted_data = utils.base64_url_decode(ed)
        conf = storage.load_configuration(url, encrypted_data)
        if isinstance(conf, dict):
            params.config.update(conf)

    for name in PROTECTED_PROPERTIES + PROTECTED_CONNECTED_PROPERTIES + PROTECTED_READONLY_PROPERTIES:
        config_name, params_name = split_name(name)
        if config_name in params.config:
            value = params.config.get(config_name) or ''
            if value and hasattr(params, params_name):
                setattr(params, params_name, value)

    for name in EDITABLE_PROPERTIES:
        config_name, params_name = split_name(name)
        if hasattr(params, params_name):
            value = params.config.get(config_name) or ''
            setattr(params, params_name, value)

    for name in BOOL_PROPERTIES:
        config_name, params_name = split_name(name)
        if hasattr(params, params_name):
            value = params.config.get(config_name)
            if isinstance(value, bool):
                setattr(params, params_name, value)

    for name in INT_PROPERTIES:
        config_name, params_name = split_name(name)
        if hasattr(params, params_name):
            value = params.config.get(config_name)
            if isinstance(value, int):
                setattr(params, params_name, value)



import sys
if sys.version_info >= (3, 8):
    import unittest
    from unittest.mock import patch
    import socket
    from datetime import timedelta
    from keepercommander.service.config.config_validation import ConfigValidator
    from keepercommander.service.util.exceptions import ValidationError

    class TestConfigValidator(unittest.TestCase):
        def setUp(self):
            self.validator = ConfigValidator()

        def test_validate_port_valid(self):
            """Test port validation with valid port numbers"""
            test_ports = [80, 443, 8080, 1024, 65535]
            for port in test_ports:
                with self.subTest(port=port):
                    with patch('socket.socket') as mock_socket:
                        mock_socket.return_value.__enter__.return_value.bind.return_value = None
                        result = self.validator.validate_port(port)
                        self.assertEqual(result, port)

        def test_validate_port_invalid_number(self):
            """Test port validation with invalid port numbers"""
            invalid_ports = [-1, 65536, 'abc', '']
            for port in invalid_ports:
                with self.subTest(port=port):
                    with self.assertRaises(ValidationError):
                        self.validator.validate_port(port)

        def test_validate_port_in_use(self):
            """Test port validation when port is already in use"""
            with patch('socket.socket') as mock_socket:
                mock_socket.return_value.__enter__.return_value.bind.side_effect = socket.error()
                with self.assertRaises(ValidationError) as context:
                    self.validator.validate_port(8080)
                self.assertIn("is already in use", str(context.exception))

        def test_validate_ngrok_token_valid(self):
            """Test ngrok token validation with valid tokens"""
            valid_tokens = [
                '1234567890abcdef',
                'abcdef1234567890',
                'abc123_def456-789'
            ]
            for token in valid_tokens:
                with self.subTest(token=token):
                    result = self.validator.validate_ngrok_token(token)
                    self.assertEqual(result, token)

        def test_validate_ngrok_token_invalid(self):
            """Test ngrok token validation with invalid tokens"""
            invalid_tokens = [
                '',
                '123',
                'abc@def',
                None
            ]
            for token in invalid_tokens:
                with self.subTest(token=token):
                    with self.assertRaises(ValidationError):
                        self.validator.validate_ngrok_token(token)

        def test_validate_rate_limit_valid(self):
            """Test rate limit validation with valid formats"""
            valid_limits = [
                '10/minute',
                '100/hour',
                '1000/day',
                '50 per minute',
                '200 per hour',
                '5000 per day'
            ]
            for limit in valid_limits:
                with self.subTest(limit=limit):
                    result = self.validator.validate_rate_limit(limit)
                    self.assertEqual(result, limit)

        def test_validate_rate_limit_invalid(self):
            """Test rate limit validation with invalid formats"""
            invalid_limits = [
                'abc',
                '10/second',
                '100 by hour',
            ]
            for limit in invalid_limits:
                with self.subTest(limit=limit):
                    with self.assertRaises(ValidationError):
                        self.validator.validate_rate_limit(limit)

        def test_validate_ip_list_valid(self):
            """Test IP list validation with valid IPs and CIDR blocks"""
            valid_ips = [
                '192.168.1.1',
                '10.0.0.0/24',
                '192.168.1.1,10.0.0.0/24',
                '2001:db8::1',
                'fe80::/10'
            ]
            for ip_list in valid_ips:
                with self.subTest(ip_list=ip_list):
                    result = self.validator.validate_ip_list(ip_list)
                    self.assertEqual(result, ip_list)

        def test_validate_ip_list_invalid(self):
            """Test IP list validation with invalid IPs"""
            invalid_ips = [
                '256.256.256.256',
                '192.168.1',
                '2001:xyz::1',
                '192.168.1.1/33',
            ]
            for ip_list in invalid_ips:
                with self.subTest(ip_list=ip_list):
                    with self.assertRaises(ValidationError):
                        self.validator.validate_ip_list(ip_list)

        def test_validate_encryption_key_valid(self):
            """Test encryption key validation with valid keys"""
            valid_key = 'abcdef1234567890ABCDEF1234567890'
            result = self.validator.validate_encryption_key(valid_key)
            self.assertEqual(result, valid_key)

        def test_validate_encryption_key_invalid(self):
            """Test encryption key validation with invalid keys"""
            invalid_keys = [
                '',
                '123456',
                'a' * 31,
                'a' * 33,
                'abc$%^&*()',
                None
            ]
            for key in invalid_keys:
                with self.subTest(key=key):
                    with self.assertRaises(ValidationError):
                        self.validator.validate_encryption_key(key)

        def test_parse_expiration_time_valid(self):
            """Test expiration time parsing with valid formats"""
            test_cases = [
                ('30m', timedelta(minutes=30)),
                ('24h', timedelta(hours=24)),
                ('7d', timedelta(days=7))
            ]
            for input_str, expected in test_cases:
                with self.subTest(input_str=input_str):
                    result = self.validator.parse_expiration_time(input_str)
                    self.assertEqual(result, expected)

        def test_parse_expiration_time_invalid(self):
            """Test expiration time parsing with invalid formats"""
            invalid_times = [
                '',
                '30x',
                '-30m',
                '0m',
                'abc',
            ]
            for time_str in invalid_times:
                with self.subTest(time_str=time_str):
                    with self.assertRaises(ValidationError):
                        self.validator.parse_expiration_time(time_str)


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from typing import Dict, Any
from configparser import ConfigParser
from pathlib import Path
from ..config.service_config import ServiceConfig
from ..decorators.logging import logger, debug_decorator
from ..util.exceptions import ValidationError
from ... import resources
from ...params import KeeperParams

class ServiceConfigHandler:
    def __init__(self, service_config: ServiceConfig):
        self.service_config = service_config
        self.config = ConfigParser()
        config_path = Path(resources.__file__).parent / 'service_config.ini'
        self.config.read(config_path)
        self.messages = self.config['Messages']
        self.validation_messages = self.config['Validation_Messages']

    def _get_validated_input(self, prompt_key: str, validation_func, error_key: str, required: bool = False):
        """Get and validate user input with consistent error handling."""
        while True:
            try:
                value = input(self.messages.get(prompt_key, '')).strip()
                if required and not value:
                    print(self.validation_messages.get(f'{error_key}_required', 'This field is required.'))
                    continue
                return validation_func(value) if value else ""
            except ValidationError as e:
                print(f"{self.validation_messages.get(error_key, 'Invalid input')} {str(e)}")
            except (KeyboardInterrupt, EOFError):
                print("\nInput cancelled by user.")
                raise
            except Exception as e:
                print(f"Unexpected error: {str(e)}")

    @debug_decorator
    def handle_streamlined_config(self, config_data: Dict[str, Any], args, params: KeeperParams) -> None:
        if args.allowedip is None:
            args.allowedip = '0.0.0.0/0,::/0'
        
        run_mode = args.run_mode if args.run_mode is not None else "foreground"
        if args.run_mode is not None and run_mode not in ['foreground', 'background']:
            raise ValidationError(f"Invalid run mode '{run_mode}'. Must be 'foreground' or 'background'.")
        
        if args.fileformat is not None and args.fileformat not in ['json', 'yaml']:
            raise ValidationError(f"Invalid file format '{args.fileformat}'. Must be 'json' or 'yaml'.")
        
        queue_enabled = args.queue_enabled if args.queue_enabled is not None else "y"
        if args.queue_enabled is not None and queue_enabled not in ['y', 'n']:
            raise ValidationError(f"Invalid queue setting '{queue_enabled}'. Must be 'y' or 'n'.")
        
        # Apply logical tunneling flow for streamlined config
        ngrok_enabled = "y" if args.ngrok else "n"
        cloudflare_enabled = "y" if args.cloudflare else "n"
        
        # Implement the same logic as interactive mode
        if ngrok_enabled == "y":
            # ngrok enabled → disable cloudflare and TLS
            cloudflare_enabled = "n"
            cloudflare_token = ""
            cloudflare_domain = ""
            tls_enabled = "n"
            certfile = ""
            certpassword = ""
            logger.debug("Ngrok enabled - disabling cloudflare and TLS")
        elif cloudflare_enabled == "y":
            # cloudflare enabled → disable TLS, but validate required fields
            if not args.cloudflare:
                raise ValidationError("Cloudflare tunnel token is required when using Cloudflare tunnel.")
            if not args.cloudflare_custom_domain:
                raise ValidationError("Cloudflare custom domain is required when using Cloudflare tunnel.")
            
            tls_enabled = "n"
            certfile = ""
            certpassword = ""
            cloudflare_token = self.service_config.validator.validate_cloudflare_token(args.cloudflare)
            cloudflare_domain = self.service_config.validator.validate_domain(args.cloudflare_custom_domain)
            logger.debug("Cloudflare enabled - disabling TLS")
        else:
            # Both ngrok and cloudflare disabled → allow TLS
            tls_enabled = "y" if args.certfile and args.certpassword else "n"
            certfile = args.certfile if args.certfile else ""
            certpassword = args.certpassword if args.certpassword else ""
            cloudflare_token = ""
            cloudflare_domain = ""
            logger.debug("No tunnels enabled - TLS configuration allowed")

        config_data.update({
            "port": self.service_config.validator.validate_port(args.port),
            "ip_allowed_list": self.service_config.validator.validate_ip_list(args.allowedip),
            "ip_denied_list": self.service_config.validator.validate_ip_list(args.deniedip),
            "ngrok": ngrok_enabled,
            "ngrok_auth_token": (
                self.service_config.validator.validate_ngrok_token(args.ngrok)
                if ngrok_enabled == "y" else ""
            ),
            "ngrok_custom_domain": args.ngrok_custom_domain if ngrok_enabled == "y" else "",
            "cloudflare": cloudflare_enabled,
            "cloudflare_tunnel_token": cloudflare_token,
            "cloudflare_custom_domain": cloudflare_domain,
            "tls_certificate": tls_enabled,
            "certfile": certfile,
            "certpassword": certpassword,
            "fileformat": args.fileformat,  # Keep original logic - can be None
            "run_mode": run_mode,
            "queue_enabled": queue_enabled
        })

    @debug_decorator
    def handle_interactive_config(self, config_data: Dict[str, Any], params: KeeperParams) -> None:
        self._configure_port(config_data)
        self._configure_tunneling_and_tls(config_data)  # New consolidated method
        self._configure_queue(config_data)
        
        config_data["fileformat"] = None
    
    def _configure_port(self, config_data: Dict[str, Any]) -> None:
        while True:
            try:
                port = input(self.messages['port_prompt'])
                logger.debug(f"Validating port: {port}")
                config_data["port"] = self.service_config.validator.validate_port(port)
                break
            except ValidationError as e:
                print(f"{self.validation_messages['invalid_port']} {str(e)}")
    
    def _configure_tunneling_and_tls(self, config_data: Dict[str, Any]) -> None:
        """
        Configure tunneling and TLS with logical flow:
        1. If ngrok = yes → Skip cloudflare and TLS (ngrok provides public access with SSL)
        2. If ngrok = no → Ask for cloudflare
        3. If ngrok = no AND cloudflare = no → Ask for TLS (local HTTPS)
        """
        # First, always ask for ngrok
        self._configure_ngrok(config_data)
        
        if config_data["ngrok"] == "y":
            # ngrok provides public access with SSL, so skip cloudflare and TLS
            config_data["cloudflare"] = "n"
            config_data["cloudflare_tunnel_token"] = ""
            config_data["cloudflare_custom_domain"] = ""
            config_data["tls_certificate"] = "n"
            config_data["certfile"] = ""
            config_data["certpassword"] = ""
        else:
            # ngrok = no, so ask for cloudflare
            self._configure_cloudflare(config_data)
            
            if config_data["cloudflare"] == "y":
                # cloudflare provides public access with SSL, so skip TLS
                config_data["tls_certificate"] = "n"
                config_data["certfile"] = ""
                config_data["certpassword"] = ""
            else:
                # Both ngrok and cloudflare = no, so ask for TLS for local HTTPS
                self._configure_tls(config_data)

    def _configure_ngrok(self, config_data: Dict[str, Any]) -> None:
        config_data["ngrok"] = self.service_config._get_yes_no_input(self.messages['ngrok_prompt'])
        
        if config_data["ngrok"] == "y":
            while True:
                try:
                    token = input(self.messages['ngrok_token_prompt'])
                    config_data["ngrok_auth_token"] = self.service_config.validator.validate_ngrok_token(token)
                    config_data["ngrok_custom_domain"]  = input(self.messages['ngrok_custom_domain_prompt'])
                    # print(f"ngrok custom domain >> "+{config_data["ngrok_custom_domain"]})
                    break
                except ValidationError as e:
                    print(f"{self.validation_messages['invalid_ngrok_token']} {str(e)}")
        else:
            config_data["ngrok_auth_token"] = ""

    def _configure_cloudflare(self, config_data: Dict[str, Any]) -> None:
        config_data["cloudflare"] = self.service_config._get_yes_no_input(
            self.messages.get('cloudflare_prompt', 'Do you want to use Cloudflare tunnel? (y/n): ')
        )
        
        if config_data["cloudflare"] == "y":
            config_data["cloudflare_tunnel_token"] = self._get_validated_input(
                prompt_key='cloudflare_token_prompt',
                validation_func=self.service_config.validator.validate_cloudflare_token,
                error_key='invalid_cloudflare_token',
                required=True
            )
            
            config_data["cloudflare_custom_domain"] = self._get_validated_input(
                prompt_key='cloudflare_custom_domain_prompt', 
                validation_func=self.service_config.validator.validate_domain,
                error_key='invalid_cloudflare_domain',
                required=True
            )
        else:
            config_data["cloudflare_tunnel_token"] = ""
            config_data["cloudflare_custom_domain"] = ""
        
    def _configure_tls(self, config_data: Dict[str, Any]) -> None:
        config_data["tls_certificate"] = self.service_config._get_yes_no_input(self.messages['tls_certificate'])
        
        if config_data["tls_certificate"] == "y":
            while True:
                try:
                    certfile = input(self.messages['certfile'])
                    certpassword = input(self.messages['certpassword'])
                    config_data["certfile"] = self.service_config.validator.validate_cert_file(certfile)
                    config_data["certpassword"]  = self.service_config.validator.validate_certpassword(certpassword)
                    # print(f"ngrok custom domain >> "+{config_data["ngrok_custom_domain"]})
                    break
                except ValidationError as e:
                    print(f"{self.validation_messages['invalid_certificate']} {str(e)}")
        else:
            config_data["certfile"] = ""
            config_data["certpassword"] = ""
    
    def _configure_queue(self, config_data: Dict[str, Any]) -> None:
        """Configure queue enabled setting with user prompt."""
        config_data["queue_enabled"] = self.service_config._get_yes_no_input(self.messages['queue_enabled_prompt'])
        logger.debug("Queue configuration completed")

    def _configure_run_mode(self, config_data: Dict[str, Any]) -> None:
        """Configure run mode with user prompt."""
        while True:
            run_mode = input(self.messages['run_mode_prompt']).strip().lower()
            if run_mode in ['foreground', 'background']:
                config_data["run_mode"] = run_mode
                logger.debug(f"Run mode set to: {run_mode}")
                break
            print(f"{self.validation_messages['invalid_run_mode']} Must be 'foreground' or 'background'.")



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: commander@keepersecurity.com
#
# Example code to create a Team and add users to the team
#
# This example also pulls configuration from config.json 
# or writes the config file if it does not exist.
#
# Usage:
#    python3 create_mc.py

import getpass
import os

from keepercommander import api
from keepercommander.__main__ import get_params_from_config
from keepercommander.commands import msp

my_params = get_params_from_config(os.path.join(os.path.dirname(__file__), 'config.json'))
while not my_params.user:
    my_params.user = getpass.getpass(prompt='User(Email): ', stream=None)

api.login(my_params)
if not my_params.session_token:
    exit(1)

api.query_enterprise(my_params)

# create MC
# msp_add_parser.add_argument('--node', dest='node', action='store', help='node name or node ID')
# msp_add_parser.add_argument('-s', '--seats', dest='seats', action='store', type=int,
#                             help='Maximum licences allowed. -1: unlimited')
# msp_add_parser.add_argument('-p', '--plan', dest='plan', action='store', required=True,
#                             choices=['business', 'businessPlus', 'enterprise', 'enterprisePlus'])
# msp_add_parser.add_argument('-f', '--file-plan', dest='file_plan', action='store',
#                             choices=['100gb', '1tb', '10tb'])
# msp_add_parser.add_argument('-a', '--addon', dest='addon', action='append', metavar='ADDON[:SEATS]',
#                             help=f'Add-ons: enterprise_breach_watch, compliance_report, enterprise_audit_and_reporting,
#                             msp_service_and_support, secrets_manager, connection_manager:N, chat
# msp_add_parser.add_argument('name', action='store', help='Managed Company name')

mc_add_command = msp.MSPAddCommand()
node = None              # optional. node ID or Name. parameter name is the same as "dest"
seats = -1               # optional
plan = 'businessPlus'    # required
file_plan = None         # optional.  '100gb' or '1tb' or '10tb'
addon = ['compliance_report']  # optional. should be an array of addons
# Add-ons: enterprise_breach_watch, compliance_report, enterprise_audit_and_reporting, msp_service_and_support, secrets_manager, connection_manager:N, chat
name = 'New MC'          # required

# create businessPlus MC with unlimited seats,  compliance_report addon
# MC ID is returned
mc_id = mc_add_command.execute(my_params, plan=plan, name=name, addon=addon)

# update MC
# msp_update_parser.add_argument('-p', '--plan', dest='plan', action='store',
#                                choices=['business', 'businessPlus', 'enterprise', 'enterprisePlus'])
# msp_update_parser.add_argument('-s', '--seats', dest='seats', action='store', type=int,
#                                help='Maximum licences allowed. -1: unlimited')
# msp_update_parser.add_argument('-f', '--file-plan', dest='file_plan', action='store',
#                                choices=['100gb', '1tb', '10tb'])
#
# Add-ons: enterprise_breach_watch, compliance_report, enterprise_audit_and_reporting, msp_service_and_support, secrets_manager, connection_manager:N, chat
#
# msp_update_parser.add_argument('-aa', '--add-addon', dest='add_addon', action='append', metavar='ADDON[:SEATS]'')
# msp_update_parser.add_argument('-ra', '--remove-addon', dest='remove_addon', action='append', metavar='ADDON')
# msp_update_parser.add_argument('mc', action='store',
#                                help='Managed Company identifier (name or id). Ex. 3862 OR "Keeper Security, Inc."')

mc_update_command = msp.MSPUpdateCommand()
seats = 4
add_addon = ['secrets_manager']
remove_addon = ['compliance_report']
mc_update_command.execute(my_params, mc=mc_id, seats=seats, add_addon=add_addon, remove_addon=remove_addon)


# remove MC
mc_remove_command = msp.MSPRemoveCommand()
mc_remove_command.execute(my_params, mc=mc_id, force=True)

api.query_enterprise(my_params)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pedm.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import folder_pb2 as folder__pb2
from . import NotificationCenter_pb2 as NotificationCenter__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\npedm.proto\x12\x04PEDM\x1a\x0c\x66older.proto\x1a\x18NotificationCenter.proto\"O\n\x17PEDMTOTPValidateRequest\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x14\n\x0c\x65nterpriseId\x18\x02 \x01(\x05\x12\x0c\n\x04\x63ode\x18\x03 \x01(\x05\";\n\nPedmStatus\x12\x0b\n\x03key\x18\x01 \x03(\x0c\x12\x0f\n\x07success\x18\x02 \x01(\x08\x12\x0f\n\x07message\x18\x03 \x01(\t\"\x89\x01\n\x12PedmStatusResponse\x12#\n\taddStatus\x18\x01 \x03(\x0b\x32\x10.PEDM.PedmStatus\x12&\n\x0cupdateStatus\x18\x02 \x03(\x0b\x32\x10.PEDM.PedmStatus\x12&\n\x0cremoveStatus\x18\x03 \x03(\x0b\x32\x10.PEDM.PedmStatus\"4\n\x0e\x44\x65ploymentData\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0c\x65\x63PrivateKey\x18\x02 \x01(\x0c\"\x9a\x01\n\x17\x44\x65ploymentCreateRequest\x12\x15\n\rdeploymentUid\x18\x01 \x01(\x0c\x12\x0e\n\x06\x61\x65sKey\x18\x02 \x01(\x0c\x12\x13\n\x0b\x65\x63PublicKey\x18\x03 \x01(\x0c\x12\x19\n\x11spiffeCertificate\x18\x04 \x01(\x0c\x12\x15\n\rencryptedData\x18\x05 \x01(\x0c\x12\x11\n\tagentData\x18\x06 \x01(\x0c\"\x8d\x01\n\x17\x44\x65ploymentUpdateRequest\x12\x15\n\rdeploymentUid\x18\x01 \x01(\x0c\x12\x15\n\rencryptedData\x18\x02 \x01(\x0c\x12)\n\x08\x64isabled\x18\x03 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12\x19\n\x11spiffeCertificate\x18\x04 \x01(\x0c\"\xa2\x01\n\x17ModifyDeploymentRequest\x12\x34\n\raddDeployment\x18\x01 \x03(\x0b\x32\x1d.PEDM.DeploymentCreateRequest\x12\x37\n\x10updateDeployment\x18\x02 \x03(\x0b\x32\x1d.PEDM.DeploymentUpdateRequest\x12\x18\n\x10removeDeployment\x18\x03 \x03(\x0c\"a\n\x0b\x41gentUpdate\x12\x10\n\x08\x61gentUid\x18\x01 \x01(\x0c\x12)\n\x08\x64isabled\x18\x02 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12\x15\n\rdeploymentUid\x18\x03 \x01(\x0c\"Q\n\x12ModifyAgentRequest\x12&\n\x0bupdateAgent\x18\x02 \x03(\x0b\x32\x11.PEDM.AgentUpdate\x12\x13\n\x0bremoveAgent\x18\x03 \x03(\x0c\"p\n\tPolicyAdd\x12\x11\n\tpolicyUid\x18\x01 \x01(\x0c\x12\x11\n\tplainData\x18\x02 \x01(\x0c\x12\x15\n\rencryptedData\x18\x03 \x01(\x0c\x12\x14\n\x0c\x65ncryptedKey\x18\x04 \x01(\x0c\x12\x10\n\x08\x64isabled\x18\x05 \x01(\x08\"v\n\x0cPolicyUpdate\x12\x11\n\tpolicyUid\x18\x01 \x01(\x0c\x12\x11\n\tplainData\x18\x02 \x01(\x0c\x12\x15\n\rencryptedData\x18\x03 \x01(\x0c\x12)\n\x08\x64isabled\x18\x04 \x01(\x0e\x32\x17.Folder.SetBooleanValue\"s\n\rPolicyRequest\x12\"\n\taddPolicy\x18\x01 \x03(\x0b\x32\x0f.PEDM.PolicyAdd\x12(\n\x0cupdatePolicy\x18\x02 \x03(\x0b\x32\x12.PEDM.PolicyUpdate\x12\x14\n\x0cremovePolicy\x18\x03 \x03(\x0c\"6\n\nPolicyLink\x12\x11\n\tpolicyUid\x18\x01 \x01(\x0c\x12\x15\n\rcollectionUid\x18\x02 \x03(\x0c\"E\n\x1aSetPolicyCollectionRequest\x12\'\n\rsetCollection\x18\x01 \x03(\x0b\x32\x10.PEDM.PolicyLink\"W\n\x0f\x43ollectionValue\x12\x15\n\rcollectionUid\x18\x01 \x01(\x0c\x12\x16\n\x0e\x63ollectionType\x18\x02 \x01(\x05\x12\x15\n\rencryptedData\x18\x03 \x01(\x0c\"z\n\x12\x43ollectionLinkData\x12\x15\n\rcollectionUid\x18\x01 \x01(\x0c\x12\x0f\n\x07linkUid\x18\x02 \x01(\x0c\x12*\n\x08linkType\x18\x03 \x01(\x0e\x32\x18.PEDM.CollectionLinkType\x12\x10\n\x08linkData\x18\x04 \x01(\x0c\"\x8c\x01\n\x11\x43ollectionRequest\x12,\n\raddCollection\x18\x01 \x03(\x0b\x32\x15.PEDM.CollectionValue\x12/\n\x10updateCollection\x18\x02 \x03(\x0b\x32\x15.PEDM.CollectionValue\x12\x18\n\x10removeCollection\x18\x03 \x03(\x0c\"{\n\x18SetCollectionLinkRequest\x12/\n\raddCollection\x18\x01 \x03(\x0b\x32\x18.PEDM.CollectionLinkData\x12.\n\x10removeCollection\x18\x02 \x03(\x0b\x32\x14.PEDM.CollectionLink\"F\n\x15\x41pprovalActionRequest\x12\x0f\n\x07\x61pprove\x18\x01 \x03(\x0c\x12\x0c\n\x04\x64\x65ny\x18\x02 \x03(\x0c\x12\x0e\n\x06remove\x18\x03 \x03(\x0c\"\xab\x01\n\x0e\x44\x65ploymentNode\x12\x15\n\rdeploymentUid\x18\x01 \x01(\x0c\x12\x10\n\x08\x64isabled\x18\x02 \x01(\x08\x12\x0e\n\x06\x61\x65sKey\x18\x03 \x01(\x0c\x12\x13\n\x0b\x65\x63PublicKey\x18\x04 \x01(\x0c\x12\x15\n\rencryptedData\x18\x05 \x01(\x0c\x12\x11\n\tagentData\x18\x06 \x01(\x0c\x12\x0f\n\x07\x63reated\x18\x07 \x01(\x03\x12\x10\n\x08modified\x18\x08 \x01(\x03\"\xa8\x01\n\tAgentNode\x12\x10\n\x08\x61gentUid\x18\x01 \x01(\x0c\x12\x11\n\tmachineId\x18\x02 \x01(\t\x12\x15\n\rdeploymentUid\x18\x03 \x01(\x0c\x12\x13\n\x0b\x65\x63PublicKey\x18\x04 \x01(\x0c\x12\x10\n\x08\x64isabled\x18\x05 \x01(\x08\x12\x15\n\rencryptedData\x18\x06 \x01(\x0c\x12\x0f\n\x07\x63reated\x18\x07 \x01(\x03\x12\x10\n\x08modified\x18\x08 \x01(\x03\"\x94\x01\n\nPolicyNode\x12\x11\n\tpolicyUid\x18\x01 \x01(\x0c\x12\x11\n\tplainData\x18\x02 \x01(\x0c\x12\x15\n\rencryptedData\x18\x03 \x01(\x0c\x12\x14\n\x0c\x65ncryptedKey\x18\x04 \x01(\x0c\x12\x0f\n\x07\x63reated\x18\x05 \x01(\x03\x12\x10\n\x08modified\x18\x06 \x01(\x03\x12\x10\n\x08\x64isabled\x18\x07 \x01(\x08\"g\n\x0e\x43ollectionNode\x12\x15\n\rcollectionUid\x18\x01 \x01(\x0c\x12\x16\n\x0e\x63ollectionType\x18\x02 \x01(\x05\x12\x15\n\rencryptedData\x18\x03 \x01(\x0c\x12\x0f\n\x07\x63reated\x18\x04 \x01(\x03\"d\n\x0e\x43ollectionLink\x12\x15\n\rcollectionUid\x18\x01 \x01(\x0c\x12\x0f\n\x07linkUid\x18\x02 \x01(\x0c\x12*\n\x08linkType\x18\x03 \x01(\x0e\x32\x18.PEDM.CollectionLinkType\"\x9d\x01\n\x12\x41pprovalStatusNode\x12\x13\n\x0b\x61pprovalUid\x18\x01 \x01(\x0c\x12\x46\n\x0e\x61pprovalStatus\x18\x02 \x01(\x0e\x32..NotificationCenter.NotificationApprovalStatus\x12\x18\n\x10\x65nterpriseUserId\x18\x03 \x01(\x03\x12\x10\n\x08modified\x18\n \x01(\x03\"\xb3\x01\n\x0c\x41pprovalNode\x12\x13\n\x0b\x61pprovalUid\x18\x01 \x01(\x0c\x12\x14\n\x0c\x61pprovalType\x18\x02 \x01(\x05\x12\x10\n\x08\x61gentUid\x18\x03 \x01(\x0c\x12\x13\n\x0b\x61\x63\x63ountInfo\x18\x04 \x01(\x0c\x12\x17\n\x0f\x61pplicationInfo\x18\x05 \x01(\x0c\x12\x15\n\rjustification\x18\x06 \x01(\x0c\x12\x10\n\x08\x65xpireIn\x18\x07 \x01(\x05\x12\x0f\n\x07\x63reated\x18\n \x01(\x03\"C\n\rFullSyncToken\x12\x15\n\rstartRevision\x18\x01 \x01(\x03\x12\x0e\n\x06\x65ntity\x18\x02 \x01(\x05\x12\x0b\n\x03key\x18\x03 \x03(\x0c\"$\n\x0cIncSyncToken\x12\x14\n\x0clastRevision\x18\x02 \x01(\x03\"h\n\rPedmSyncToken\x12\'\n\x08\x66ullSync\x18\x02 \x01(\x0b\x32\x13.PEDM.FullSyncTokenH\x00\x12%\n\x07incSync\x18\x03 \x01(\x0b\x32\x12.PEDM.IncSyncTokenH\x00\x42\x07\n\x05token\"/\n\x12GetPedmDataRequest\x12\x19\n\x11\x63ontinuationToken\x18\x01 \x01(\x0c\"\xad\x04\n\x13GetPedmDataResponse\x12\x19\n\x11\x63ontinuationToken\x18\x01 \x01(\x0c\x12\x12\n\nresetCache\x18\x02 \x01(\x08\x12\x0f\n\x07hasMore\x18\x03 \x01(\x08\x12\x1a\n\x12removedDeployments\x18\n \x03(\x0c\x12\x15\n\rremovedAgents\x18\x0b \x03(\x0c\x12\x17\n\x0fremovedPolicies\x18\x0c \x03(\x0c\x12\x19\n\x11removedCollection\x18\r \x03(\x0c\x12\x33\n\x15removedCollectionLink\x18\x0e \x03(\x0b\x32\x14.PEDM.CollectionLink\x12\x18\n\x10removedApprovals\x18\x0f \x03(\x0c\x12)\n\x0b\x64\x65ployments\x18\x14 \x03(\x0b\x32\x14.PEDM.DeploymentNode\x12\x1f\n\x06\x61gents\x18\x15 \x03(\x0b\x32\x0f.PEDM.AgentNode\x12\"\n\x08policies\x18\x16 \x03(\x0b\x32\x10.PEDM.PolicyNode\x12)\n\x0b\x63ollections\x18\x17 \x03(\x0b\x32\x14.PEDM.CollectionNode\x12,\n\x0e\x63ollectionLink\x18\x18 \x03(\x0b\x32\x14.PEDM.CollectionLink\x12%\n\tapprovals\x18\x19 \x03(\x0b\x32\x12.PEDM.ApprovalNode\x12\x30\n\x0e\x61pprovalStatus\x18\x1a \x03(\x0b\x32\x18.PEDM.ApprovalStatusNode\"<\n\x12PolicyAgentRequest\x12\x11\n\tpolicyUid\x18\x01 \x03(\x0c\x12\x13\n\x0bsummaryOnly\x18\x02 \x01(\x08\";\n\x13PolicyAgentResponse\x12\x12\n\nagentCount\x18\x01 \x01(\x05\x12\x10\n\x08\x61gentUid\x18\x02 \x03(\x0c\"]\n\x16\x41uditCollectionRequest\x12\x19\n\x11\x63ontinuationToken\x18\x01 \x01(\x0c\x12\x10\n\x08valueUid\x18\x02 \x03(\x0c\x12\x16\n\x0e\x63ollectionName\x18\x03 \x03(\t\"h\n\x14\x41uditCollectionValue\x12\x16\n\x0e\x63ollectionName\x18\x01 \x01(\t\x12\x10\n\x08valueUid\x18\x02 \x01(\x0c\x12\x15\n\rencryptedData\x18\x03 \x01(\x0c\x12\x0f\n\x07\x63reated\x18\x04 \x01(\x03\"q\n\x17\x41uditCollectionResponse\x12*\n\x06values\x18\x01 \x03(\x0b\x32\x1a.PEDM.AuditCollectionValue\x12\x0f\n\x07hasMore\x18\x02 \x01(\x08\x12\x19\n\x11\x63ontinuationToken\x18\x03 \x01(\x0c\"H\n\x18GetCollectionLinkRequest\x12,\n\x0e\x63ollectionLink\x18\x01 \x03(\x0b\x32\x14.PEDM.CollectionLink\"Q\n\x19GetCollectionLinkResponse\x12\x34\n\x12\x63ollectionLinkData\x18\x01 \x03(\x0b\x32\x18.PEDM.CollectionLinkData\"2\n\x1aGetActiveAgentCountRequest\x12\x14\n\x0c\x65nterpriseId\x18\x01 \x03(\x05\">\n\x10\x41\x63tiveAgentCount\x12\x14\n\x0c\x65nterpriseId\x18\x01 \x01(\x05\x12\x14\n\x0c\x61\x63tiveAgents\x18\x02 \x01(\x05\";\n\x12\x41\x63tiveAgentFailure\x12\x14\n\x0c\x65nterpriseId\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\"x\n\x1bGetActiveAgentCountResponse\x12*\n\nagentCount\x18\x01 \x03(\x0b\x32\x16.PEDM.ActiveAgentCount\x12-\n\x0b\x66\x61iledCount\x18\x02 \x03(\x0b\x32\x18.PEDM.ActiveAgentFailure\"\x87\x01\n\x19GetAgentDailyCountRequest\x12\x14\n\x0c\x65nterpriseId\x18\x01 \x03(\x05\x12$\n\tmonthYear\x18\x02 \x01(\x0b\x32\x0f.PEDM.MonthYearH\x00\x12$\n\tdateRange\x18\x03 \x01(\x0b\x32\x0f.PEDM.DateRangeH\x00\x42\x08\n\x06period\"(\n\tMonthYear\x12\r\n\x05month\x18\x01 \x01(\x05\x12\x0c\n\x04year\x18\x02 \x01(\x05\"\'\n\tDateRange\x12\r\n\x05start\x18\x01 \x01(\x03\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x03\"3\n\x0f\x41gentDailyCount\x12\x0c\n\x04\x64\x61te\x18\x01 \x01(\x03\x12\x12\n\nagentCount\x18\x02 \x01(\x05\"V\n\x17\x41gentCountForEnterprise\x12\x14\n\x0c\x65nterpriseId\x18\x01 \x01(\x05\x12%\n\x06\x63ounts\x18\x02 \x03(\x0b\x32\x15.PEDM.AgentDailyCount\"U\n\x1aGetAgentDailyCountResponse\x12\x37\n\x10\x65nterpriseCounts\x18\x01 \x03(\x0b\x32\x1d.PEDM.AgentCountForEnterprise*j\n\x12\x43ollectionLinkType\x12\r\n\tCLT_OTHER\x10\x00\x12\r\n\tCLT_AGENT\x10\x01\x12\x0e\n\nCLT_POLICY\x10\x02\x12\x12\n\x0e\x43LT_COLLECTION\x10\x03\x12\x12\n\x0e\x43LT_DEPLOYMENT\x10\x04\x42 \n\x18\x63om.keepersecurity.protoB\x04PEDMb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pedm_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\004PEDM'
  _globals['_COLLECTIONLINKTYPE']._serialized_start=5286
  _globals['_COLLECTIONLINKTYPE']._serialized_end=5392
  _globals['_PEDMTOTPVALIDATEREQUEST']._serialized_start=60
  _globals['_PEDMTOTPVALIDATEREQUEST']._serialized_end=139
  _globals['_PEDMSTATUS']._serialized_start=141
  _globals['_PEDMSTATUS']._serialized_end=200
  _globals['_PEDMSTATUSRESPONSE']._serialized_start=203
  _globals['_PEDMSTATUSRESPONSE']._serialized_end=340
  _globals['_DEPLOYMENTDATA']._serialized_start=342
  _globals['_DEPLOYMENTDATA']._serialized_end=394
  _globals['_DEPLOYMENTCREATEREQUEST']._serialized_start=397
  _globals['_DEPLOYMENTCREATEREQUEST']._serialized_end=551
  _globals['_DEPLOYMENTUPDATEREQUEST']._serialized_start=554
  _globals['_DEPLOYMENTUPDATEREQUEST']._serialized_end=695
  _globals['_MODIFYDEPLOYMENTREQUEST']._serialized_start=698
  _globals['_MODIFYDEPLOYMENTREQUEST']._serialized_end=860
  _globals['_AGENTUPDATE']._serialized_start=862
  _globals['_AGENTUPDATE']._serialized_end=959
  _globals['_MODIFYAGENTREQUEST']._serialized_start=961
  _globals['_MODIFYAGENTREQUEST']._serialized_end=1042
  _globals['_POLICYADD']._serialized_start=1044
  _globals['_POLICYADD']._serialized_end=1156
  _globals['_POLICYUPDATE']._serialized_start=1158
  _globals['_POLICYUPDATE']._serialized_end=1276
  _globals['_POLICYREQUEST']._serialized_start=1278
  _globals['_POLICYREQUEST']._serialized_end=1393
  _globals['_POLICYLINK']._serialized_start=1395
  _globals['_POLICYLINK']._serialized_end=1449
  _globals['_SETPOLICYCOLLECTIONREQUEST']._serialized_start=1451
  _globals['_SETPOLICYCOLLECTIONREQUEST']._serialized_end=1520
  _globals['_COLLECTIONVALUE']._serialized_start=1522
  _globals['_COLLECTIONVALUE']._serialized_end=1609
  _globals['_COLLECTIONLINKDATA']._serialized_start=1611
  _globals['_COLLECTIONLINKDATA']._serialized_end=1733
  _globals['_COLLECTIONREQUEST']._serialized_start=1736
  _globals['_COLLECTIONREQUEST']._serialized_end=1876
  _globals['_SETCOLLECTIONLINKREQUEST']._serialized_start=1878
  _globals['_SETCOLLECTIONLINKREQUEST']._serialized_end=2001
  _globals['_APPROVALACTIONREQUEST']._serialized_start=2003
  _globals['_APPROVALACTIONREQUEST']._serialized_end=2073
  _globals['_DEPLOYMENTNODE']._serialized_start=2076
  _globals['_DEPLOYMENTNODE']._serialized_end=2247
  _globals['_AGENTNODE']._serialized_start=2250
  _globals['_AGENTNODE']._serialized_end=2418
  _globals['_POLICYNODE']._serialized_start=2421
  _globals['_POLICYNODE']._serialized_end=2569
  _globals['_COLLECTIONNODE']._serialized_start=2571
  _globals['_COLLECTIONNODE']._serialized_end=2674
  _globals['_COLLECTIONLINK']._serialized_start=2676
  _globals['_COLLECTIONLINK']._serialized_end=2776
  _globals['_APPROVALSTATUSNODE']._serialized_start=2779
  _globals['_APPROVALSTATUSNODE']._serialized_end=2936
  _globals['_APPROVALNODE']._serialized_start=2939
  _globals['_APPROVALNODE']._serialized_end=3118
  _globals['_FULLSYNCTOKEN']._serialized_start=3120
  _globals['_FULLSYNCTOKEN']._serialized_end=3187
  _globals['_INCSYNCTOKEN']._serialized_start=3189
  _globals['_INCSYNCTOKEN']._serialized_end=3225
  _globals['_PEDMSYNCTOKEN']._serialized_start=3227
  _globals['_PEDMSYNCTOKEN']._serialized_end=3331
  _globals['_GETPEDMDATAREQUEST']._serialized_start=3333
  _globals['_GETPEDMDATAREQUEST']._serialized_end=3380
  _globals['_GETPEDMDATARESPONSE']._serialized_start=3383
  _globals['_GETPEDMDATARESPONSE']._serialized_end=3940
  _globals['_POLICYAGENTREQUEST']._serialized_start=3942
  _globals['_POLICYAGENTREQUEST']._serialized_end=4002
  _globals['_POLICYAGENTRESPONSE']._serialized_start=4004
  _globals['_POLICYAGENTRESPONSE']._serialized_end=4063
  _globals['_AUDITCOLLECTIONREQUEST']._serialized_start=4065
  _globals['_AUDITCOLLECTIONREQUEST']._serialized_end=4158
  _globals['_AUDITCOLLECTIONVALUE']._serialized_start=4160
  _globals['_AUDITCOLLECTIONVALUE']._serialized_end=4264
  _globals['_AUDITCOLLECTIONRESPONSE']._serialized_start=4266
  _globals['_AUDITCOLLECTIONRESPONSE']._serialized_end=4379
  _globals['_GETCOLLECTIONLINKREQUEST']._serialized_start=4381
  _globals['_GETCOLLECTIONLINKREQUEST']._serialized_end=4453
  _globals['_GETCOLLECTIONLINKRESPONSE']._serialized_start=4455
  _globals['_GETCOLLECTIONLINKRESPONSE']._serialized_end=4536
  _globals['_GETACTIVEAGENTCOUNTREQUEST']._serialized_start=4538
  _globals['_GETACTIVEAGENTCOUNTREQUEST']._serialized_end=4588
  _globals['_ACTIVEAGENTCOUNT']._serialized_start=4590
  _globals['_ACTIVEAGENTCOUNT']._serialized_end=4652
  _globals['_ACTIVEAGENTFAILURE']._serialized_start=4654
  _globals['_ACTIVEAGENTFAILURE']._serialized_end=4713
  _globals['_GETACTIVEAGENTCOUNTRESPONSE']._serialized_start=4715
  _globals['_GETACTIVEAGENTCOUNTRESPONSE']._serialized_end=4835
  _globals['_GETAGENTDAILYCOUNTREQUEST']._serialized_start=4838
  _globals['_GETAGENTDAILYCOUNTREQUEST']._serialized_end=4973
  _globals['_MONTHYEAR']._serialized_start=4975
  _globals['_MONTHYEAR']._serialized_end=5015
  _globals['_DATERANGE']._serialized_start=5017
  _globals['_DATERANGE']._serialized_end=5056
  _globals['_AGENTDAILYCOUNT']._serialized_start=5058
  _globals['_AGENTDAILYCOUNT']._serialized_end=5109
  _globals['_AGENTCOUNTFORENTERPRISE']._serialized_start=5111
  _globals['_AGENTCOUNTFORENTERPRISE']._serialized_end=5197
  _globals['_GETAGENTDAILYCOUNTRESPONSE']._serialized_start=5199
  _globals['_GETAGENTDAILYCOUNTRESPONSE']._serialized_end=5284
# @@protoc_insertion_point(module_scope)



from datetime import timedelta
from unittest import TestCase

import pytest

from keepercommander.commands.helpers.timeout import (
    enforce_timeout_range, format_timeout, get_delta_from_timeout_setting, get_timeout_setting_from_delta, parse_timeout
)
from keepercommander.constants import TIMEOUT_DEFAULT


@pytest.mark.parametrize(
    'time_input,expected',
    [
        ('3', timedelta(minutes=3)),
        ('1d2h3mi', timedelta(days=1, hours=2, minutes=3)),
        ('4 minutes 5 hours 6 days', timedelta(days=6, hours=5, minutes=4))
    ]
)
def test_parse_timeout(time_input, expected):
    assert parse_timeout(time_input) == expected


@pytest.mark.parametrize(
    'timeout_delta,expected',
    [
        (timedelta(), '0'),
        (timedelta(minutes=3), '3 minutes'),
        (timedelta(days=1, hours=2, minutes=3), '1 day, 2 hours, 3 minutes'),
        (timedelta(days=6, hours=5, minutes=4), '6 days, 5 hours, 4 minutes')
    ]
)
def test_format_timeout(timeout_delta, expected):
    assert format_timeout(timeout_delta) == expected


class TestParseTimeout(TestCase):
    def test_parse_timeout_invalid(self):
        time_input = '3 invalidunits'
        with self.assertRaises(ValueError):
            parse_timeout(time_input)


class TestEnforceTimeoutRange(TestCase):
    def test_less_than_min(self):
        timeout_delta = timedelta(0)
        expect_delta = TIMEOUT_DEFAULT
        new_timeout_delta = enforce_timeout_range(timeout_delta)
        self.assertEqual(expect_delta, new_timeout_delta)


class TestTimeoutSetting(TestCase):
    def test_get_timeout_setting_from_delta(self):
        timeout_delta = timedelta(minutes=20000)
        expect_setting = '20000'

        self.assertEqual(expect_setting, get_timeout_setting_from_delta(timeout_delta))

    def test_get_delta_from_timeout_setting(self):
        # 10000 (minutes) * 60000 (milliseconds per minute)
        timeout_setting = '600000000'
        expect_delta = timedelta(minutes=10000)

        self.assertEqual(expect_delta, get_delta_from_timeout_setting(timeout_setting))



from . import ConnectionBase
from ..types import DataPayload, SyncData, EdgeType
import json
import os
import logging
from tabulate import tabulate

try:  # pragma: no cover
    import sqlite3
    from contextlib import closing
except ImportError:
    raise Exception("Please install the sqlite3 module to use the Local connection.")

from typing import Optional, Union, Any, TYPE_CHECKING  # pragma: no cover
if TYPE_CHECKING:
    Logger = Union[logging.RootLogger, logging.Logger]


class Connection(ConnectionBase):

    """
    BIG TIME NOTE

    This is a fake DAG engine used for unit tests.
    It tries best to emulate krouter/workflow.
    This is no substitute for testing against a krouter instance.
    """

    DB_FILE = "local_dag.db"
    DEBUG = 0

    def __init__(self, limit: int = 100, db_file: Optional[str] = None, db_dir:  Optional[str] = None,
                 logger: Optional[Any] = None):

        super().__init__(is_device=True, logger=logger)

        if db_file is None:
            db_file = os.environ.get("LOCAL_DAG_DB_FILE", Connection.DB_FILE)
        if db_dir is None:
            db_dir = os.environ.get("LOCAL_DAG_DIR", os.environ.get("HOME", os.environ.get("USERPROFILE", "./")))

        self.db_file = os.path.join(db_dir, db_file)
        self.limit = limit

        self.create_database()

    @staticmethod
    def debug(msg):
        if Connection.DEBUG == 1:
            logging.debug(f"DAG: {msg}")

    @staticmethod
    def get_record_uid(record: object) -> bytes:
        if hasattr(record, "record_uid") is True:
            return getattr(record, "record_uid")
        elif hasattr(record, "uid") is True:
            return getattr(record, "uid")
        raise Exception(f"Cannot find the record uid in object type: {type(record)}.")

    @staticmethod
    def get_key_bytes(record: object) -> bytes:
        if hasattr(record, "record_key_bytes") is True:
            return getattr(record, "record_key_bytes")
        elif hasattr(record, "record_key") is True:
            return getattr(record, "record_key")
        raise Exception("Cannot find the record key bytes in object.")

    def clear_database(self):
        try:
            os.unlink(self.db_file)
        except (Exception,):
            pass

    def create_database(self):

        self.debug("create local dag database")

        if os.path.isfile(self.db_file) is True:
            return False

        with closing(sqlite3.connect(self.db_file)) as connection:
            with closing(connection.cursor()) as cursor:

                # This is based on workflow, Database.kt.
                # The UIDs are stored a character instead of bytes to make them more readable for debugging.

                # The 'type' columns are stored a TEXT.
                # This is because the WS wants text for the enum, but stores
                # it as an INTEGER.
                # We are just going to store it as a TEXT and avoid the middle man.

                cursor.execute(
                    """
CREATE TABLE IF NOT EXISTS dag_edges (
    graph_id INTEGER NOT NULL DEFAULT 0,
    edge_id INTEGER PRIMARY KEY AUTOINCREMENT,
    type TEXT NOT NULL,
    head CHARACTER(22) NOT NULL,
    tail CHARACTER(22) NOT NULL,
    data BLOB,
    origin CHARACTER(22),
    path TEXT,
    created timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
    creator_id BLOB(16) DEFAULT NULL,
    creator_type INTEGER DEFAULT NULL,
    creator_name TEXT DEFAULT NULL,
    FOREIGN KEY(head) REFERENCES dag_vertices(vertex_id),
    FOREIGN KEY(tail) REFERENCES dag_vertices(vertex_id)
)
                    """
                )

                cursor.execute(
                    """
CREATE TABLE IF NOT EXISTS dag_vertices (
    vertex_id CHARACTER(22) NOT NULL,
    type TEXT NOT NULL,
    name TEXT,
    owner_id BLOB(16) DEFAULT NULL
)
                    """
                )

                cursor.execute(
                    """
CREATE TABLE IF NOT EXISTS dag_streams (
    graph_id INTEGER NOT NULL,
    sync_point INTEGER PRIMARY KEY AUTOINCREMENT,
    vertex_id CHARACTER(22) NOT NULL,
    edge_id INTEGER NOT NULL,
    count INTEGER NOT NULL DEFAULT 0,
    deletion INTEGER NOT NULL DEFAULT 0,
    UNIQUE(vertex_id,edge_id),
    FOREIGN KEY(vertex_id) REFERENCES dag_vertices(vertex_id),
    FOREIGN KEY(edge_id) REFERENCES dag_edges(edge_id)
)
                    """
                )
                connection.commit()

        os.chmod(self.db_file, 0o777)

    @staticmethod
    def _payload_to_json(payload: Union[DataPayload, str]) -> str:

        # if payload is DataPayload
        if isinstance(payload, DataPayload):
            payload_data = payload.model_dump_json()
        elif isinstance(payload, str):
            payload_data = payload

            # make sure it is a valid json and raise and exception if not. make an exception for the case of a string
            # that is a valid json
            if not payload_data.startswith('{') and not payload_data.endswith('}'):
                raise Exception(f'Invalid payload: {payload_data}')

            # double check if it is a valid json inside the string
            json.loads(payload_data)
        else:
            raise Exception(f'Unsupported payload type: {type(payload)}')

        return json.loads(payload_data)

    def _find_stream_id(self, payload: DataPayload):

        data = Connection._payload_to_json(payload)

        # Find the vertex that does not belong to any other vertex.
        # This is normally root for a full DAG, but will be a vertex if adding additional edges.
        # 100% sure this could be written better.
        # 1000% sure this could be written better.
        # TODO: Only refs that are type PAM_NETWORK or PAM_USER can contain the stream id.
        #  Change code to ignore all other ref types.

        self.debug("finding stream id")

        # First check if we can route with existing edges in the database.
        stream_id = None
        with closing(sqlite3.connect(self.db_file)) as connection:
            with closing(connection.cursor()) as cursor:

                graph_id = data.get("graphId")

                stream_ids = {}

                runs = 0
                for item in data.get("dataList"):

                    # Get the head UID of the edge and then find an edge where the UID is the tail.
                    # If we find an edge, use its head to find an edge where the UID is the tail.
                    # Repeat until we can't find and edge, that is a stream ID
                    # Tally all the stream ID and take the best.
                    item_stream_id = item.get("ref")["value"]
                    current_stream_id = item_stream_id
                    while True:
                        self.debug(f"    check stream id {current_stream_id}")
                        sql = "SELECT head, edge_id FROM dag_edges WHERE tail=? AND graph_id=? AND type != ?"
                        res = cursor.execute(sql, (current_stream_id, graph_id, EdgeType.DATA.value))
                        row = res.fetchone()
                        if row is None:
                            self.debug(f"    no edge found")
                            if current_stream_id == item_stream_id:
                                current_stream_id = None
                            break
                        current_stream_id = row[0]
                        self.debug(f"      got {current_stream_id}")

                    if current_stream_id is not None:
                        if item_stream_id not in stream_ids:
                            stream_ids[current_stream_id] = 0
                        stream_ids[current_stream_id] += 1
                    else:
                        # If we didn't find anything with the tail, check starting with the head.
                        item_stream_id = item.get("parentRef")["value"]
                        current_stream_id = item_stream_id
                        while True:
                            self.debug(f"    check stream id {current_stream_id}")
                            sql = "SELECT head, edge_id FROM dag_edges WHERE tail=? AND graph_id=? AND type != ?"
                            res = cursor.execute(sql, (current_stream_id, graph_id, EdgeType.DATA.value))
                            row = res.fetchone()
                            if row is None:
                                self.debug(f"    no edge found")
                                if current_stream_id == item_stream_id:
                                    current_stream_id = None
                                break
                            current_stream_id = row[0]
                            self.debug(f"      got {current_stream_id}")

                        if current_stream_id is not None:
                            if item_stream_id not in stream_ids:
                                stream_ids[current_stream_id] = 0
                            stream_ids[current_stream_id] += 1

                    # Until we rewrite this, exit after we check 3 edges.
                    # This will slow down after a bunch of edges are added.
                    # We also fixed stuff in our code to prevent the errors we were seeing.
                    # Might want to switch to recursion.
                    # https://www.sqlite.org/lang_with.html
                    if runs > 3:
                        break
                    runs += 1

                if len(stream_ids) > 0:
                    sorted_stream_ids = [k for k, v in sorted(stream_ids.items(), key=lambda item: item[1])]
                    stream_id = sorted_stream_ids.pop()

        # If the stream id is None, this is the first save of the DAG.
        # No edges existed.
        # Compare the data list items.
        # The one without an edge with a tail if the stream id.
        if stream_id is None:
            self.debug("stream id None, edges might be new")
            # Get a starting spot
            found = {}
            for item in data.get("dataList"):
                head_uid = item.get("parentRef")["value"]
                found[head_uid] = True
            for item in data.get("dataList"):
                tail_uid = item.get("ref")["value"]
                found.pop(tail_uid, None)
            stream_ids = [uid for uid in found]
            if len(stream_ids) > 0:
                stream_id = stream_ids[0]

        # If we can't find stream ID, assume it's on the first item in the dataList
        if stream_id is None:
            item = data.get("dataList")[0]
            stream_id = item.get("parentRef")["value"] or item.get("ref")["value"]

        return stream_id

    def add_data(self, payload: DataPayload, agent: str):

        stream_id = self._find_stream_id(payload)
        self.debug(f"STREAM ID IS {stream_id}")

        data = Connection._payload_to_json(payload)

        with closing(sqlite3.connect(self.db_file)) as connection:
            with closing(connection.cursor()) as cursor:

                origin_id = data.get("origin")["value"]
                graph_id = data.get("graphId")

                saved_vertex = {}
                for item in data.get("dataList"):

                    tail_uid = item.get("ref")["value"]
                    tail_type = item.get("ref")["type"]
                    tail_name = item.get("ref")["name"]

                    head_uid = None
                    head_type = None
                    head_name = None
                    if item.get("parentRef") is not None:
                        head_uid = item.get("parentRef")["value"]
                        head_type = item.get("parentRef")["type"]
                        head_name = item.get("parentRef")["name"]

                    edge_type = item.get("type")
                    path = item.get("path")

                    sql = "INSERT INTO dag_edges (type, head, tail, data, origin, graph_id, path) "
                    sql += "VALUES (?,?,?,?,?,?,?)"
                    cursor.execute(sql, (
                        edge_type,
                        head_uid,
                        tail_uid,
                        item.get("content"),
                        origin_id,
                        graph_id,
                        path
                    ))
                    edge_id = cursor.lastrowid

                    sql = "INSERT INTO dag_streams (graph_id, vertex_id, edge_id, count) VALUES (?, ?, ?, ?)"
                    cursor.execute(sql, (
                        graph_id,
                        stream_id,
                        edge_id,
                        1
                    ))

                    if saved_vertex.get(tail_uid) is None:
                        # Type is RefType enum value
                        sql = "INSERT INTO dag_vertices (vertex_id, type, name) VALUES (?, ?, ?)"
                        cursor.execute(sql, (
                            tail_uid,
                            tail_type,
                            tail_name
                        ))
                        saved_vertex[tail_uid] = True
                    if saved_vertex.get(head_uid) is None:
                        # Type is RefType enum value
                        sql = "INSERT INTO dag_vertices (vertex_id, type, name) VALUES (?, ?, ?)"
                        cursor.execute(sql, (
                            head_uid,
                            head_type,
                            head_name
                        ))
                        saved_vertex[head_uid] = True

                connection.commit()

    def sync(self, stream_id: str,  agent: str, sync_point: Optional[int] = 0, graph_id: Optional[int] = 0) -> SyncData:

        self.debug(f"Sync: stream id {stream_id}, sync point {sync_point}, graph {graph_id}")

        edge_type_map = {
            EdgeType.DATA.value: "data",
            EdgeType.KEY.value: "key",
            EdgeType.LINK.value: "link",
            EdgeType.ACL.value: "acl",
            EdgeType.DELETION.value: "deletion",
            EdgeType.DENIAL.value: "denial",
            EdgeType.UNDENIAL.value: "undenial",
        }

        resp = {
            "syncPoint": 0,
            "data": [],
            "hasMore": False
        }

        with closing(sqlite3.connect(self.db_file)) as connection:
            with closing(connection.cursor()) as cursor:
                self.debug(f"... loading DAG, {stream_id}, {sync_point}, {self.limit + 1}")

                args = [stream_id, sync_point, graph_id]
                sql = "SELECT sync_point, edge_id FROM dag_streams WHERE vertex_id = ? AND deletion = 0 "\
                      "AND sync_point > ? AND graph_id=? ORDER BY sync_point ASC LIMIT ?"
                args.append(self.limit + 1)
                res = cursor.execute(sql, tuple(args))
                rows = list(res.fetchall())
                if len(rows) > self.limit:
                    resp["hasMore"] = True
                    rows.pop()
                for row in rows:
                    resp["syncPoint"] = row[0]

                    args = [row[1], graph_id]
                    sql = "SELECT head, tail, data, path, type FROM dag_edges WHERE edge_id = ? AND graph_id=?"
                    res = cursor.execute(sql, tuple(args))
                    edges = res.fetchone()

                    # If the head and tail are the same (DATA edge), then parent_ref is None.
                    # Else include a parent_ref
                    parent_ref = None
                    if edges[1] != edges[0]:

                        sql = "SELECT type FROM dag_vertices WHERE vertex_id = ?"
                        res = cursor.execute(sql, (edges[0],))
                        head_vertex = res.fetchone()

                        parent_ref = {
                            "type": head_vertex[0],
                            "value": edges[0],
                            "name": None
                        }

                    sql = "SELECT type FROM dag_vertices WHERE vertex_id = ?"
                    res = cursor.execute(sql, (edges[1],))
                    tail_vertex = res.fetchone()

                    resp["data"].append({
                        "type": edge_type_map.get(edges[4]),
                        "ref": {
                            "type": tail_vertex[0],
                            "value": edges[1],
                            "name": None
                        },
                        "parentRef": parent_ref,
                        "content": edges[2],
                        "path": edges[3],
                        "deletion": False
                    })

        sync_data_resp = SyncData.model_validate_json(json.dumps(resp))
        return sync_data_resp

    def debug_dump(self) -> str:

        ret = ""

        with closing(sqlite3.connect(self.db_file)) as connection:
            with closing(connection.cursor()) as cursor:

                cols = ["graph_id", "edge_id", "type", "head", "tail", "data", "origin", "path", "created",
                        "creator_id", "creator_type", "creator_name"]

                sql = f"SELECT {','.join(cols)} FROM dag_edges ORDER BY edge_id DESC"
                res = cursor.execute(sql,)

                ret += "dag_edges\n"
                ret += "=========\n"
                table = []
                for row in res.fetchall():
                    table.append(list(row))

                ret += tabulate(table, cols) + "\n\n"

                cols = ["e.graph_id", "e.edge_id", "v.vertex_id", "v.type", "v.name", "v.owner_id"]

                sql = f"SELECT {','.join(cols)} "\
                      "FROM dag_vertices v "\
                      "INNER JOIN dag_edges e ON e.tail = v.vertex_id "\
                      "ORDER BY e.graph_id DESC, e.edge_id DESC"
                res = cursor.execute(sql,)

                ret += "dag_vertices\n"
                ret += "============\n"
                table = []
                for row in res.fetchall():
                    table.append(list(row))

                ret += tabulate(table, cols) + "\n\n"

                cols = ["graph_id", "edge_id", "sync_point", "vertex_id", "count", "deletion"]

                sql = f"SELECT {','.join(cols)} FROM dag_streams ORDER BY edge_id DESC"
                res = cursor.execute(sql,)

                ret += "dag_streams\n"
                ret += "===========\n"
                table = []
                for row in res.fetchall():
                    table.append(list(row))

                ret += tabulate(table, cols) + "\n\n"

        return ret

    def update_edge_content(self, graph_id: int, head_uid: str, tail_uid: str, content: str):

        with closing(sqlite3.connect(self.db_file)) as connection:
            with closing(connection.cursor()) as cursor:

                sql = "UPDATE dag_edges SET data=? WHERE graph_id=? AND head=? AND tail=?"
                res = cursor.execute(sql, (content, graph_id, head_uid, tail_uid))

            connection.commit()

    def clear(self):

        with closing(sqlite3.connect(self.db_file)) as connection:
            with closing(connection.cursor()) as cursor:

                for table in ["dag_streams", "dag_edges", "dag_vertices"]:
                    sql = f"DELETE FROM {table}"
                    cursor.execute(sql, )

            connection.commit()
